/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include "fitsio.h"
#include "imaging_data.h"
#include "minrtsFrameData.h"
#include "minrtsFftUtil.h"
#include "cpgplot.h"
#define maxCoeff 16

/**********************************************************
 * Usage:
 *    oirFoilWaveCal datafilename darkfilename linelistfilename
 *
 *    filename is a single oir FITS table file
************************************************************** */
void Usage() {
   printf("Usage: oirFoilWaveCal [-wave foilInputFile]\n");
   printf("      [-dark darkInputFile]\n");
   printf("      [-curve curvatureInputFile]\n");
   printf("      -out outputWindowTemplateFile\n");
   printf("      [-wavedark waveCalDarkFile] (if different from common dark)\n");
   printf("      [-curvedark curveDarkFile] (if different from common dark)\n");
   printf("      [-refDet refDetectorFileName] (if different from wave/curve )\n");
   printf("      [-line foilStandardFile] \n");
   printf("      [-plot plotFile] (defaults to midiFoilPlot.ps)\n");
   printf("      [-ascii asciifile] (defaults to midiFoil.dat)\n");
   return;
};
int main (int argc, char** argv) {

   minrtsFrameData *meanData;
   minrtsFrameData *meanDark;
   minrtsFrameData *rms;
   minrtsFrameData *compressed;

   float* work;
   int ierr = 0;
   int i,j,k, iregion, nregion, npix, ipix;
   int nx, ny;

   float wt, np, x, y;
   int order;
   int dx;
   float sigma;

   FILE *lineFile;
   int nLine;
   float *xLine, *newXLine, *freqLine;
   float *xCoord;
   float *lWeight;
   float polyCoeff[maxCoeff];
   float crpix;

   fitsfile *inFile;
   fitsfile *outFile = NULL;
   char waveFileName[130];
   char waveDarkName[130];
   char curveFileName[130];
   char curveDarkName[130];
   char lineFileName[130];
   char darkFileName[130];
   char outFileName[130];
   char plotFileName[130];
   char printFileName[130];
   char refDetectorFileName[130];
   char *inFileName;
   char *bang = "!";

   char grism[6];
   imaging_data_table_def *idata, *odata;
   imaging_detector_table_def *inDetectorDef;
   imaging_detector_table_def *outDetectorDef;
   imaging_data_row *inData;
   int iKey;
   int maxstep;

   float *plotFreq[4];
   float *plotCurve[4];
   float *curveData[4];

   int iarg;
    
   FILE *printFile;

   float cLight = 2.99795e14;

/*  
       parse input
*/
   waveFileName[0]  = 0;
   waveDarkName[0]  = 0;
   curveFileName[0] = 0;
   curveDarkName[0] = 0;
   lineFileName[0]  = 0;
   outFileName[0]   = 0;
   plotFileName[0]  = 0;
   printFileName[0] = 0;
   refDetectorFileName[0] = 0;
   iarg = 1;
   while (iarg < argc ) {
      if (!strcmp("-wave", argv[iarg])) strcpy (waveFileName, argv[(iarg++)+1]);
      if (!strcmp("-refDet", argv[iarg])) strcpy (refDetectorFileName, argv[(iarg++)+1]);
      if (!strcmp("-dark", argv[iarg])) strcpy (darkFileName, argv[(iarg++)+1]);
      if (!strcmp("-wavedark", argv[iarg])) 
         strcpy (waveDarkName, argv[(iarg++)+1]);
      if (!strcmp("-curve", argv[iarg])) 
         strcpy (curveFileName, argv[(iarg++)+1]);
      if (!strcmp("-curvedark", argv[iarg])) 
         strcpy (curveDarkName, argv[(iarg++)+1]);
      if (!strcmp("-line", argv[iarg])) 
         strcpy (lineFileName, argv[(iarg++)+1]);
      if (!strcmp("-out", argv[iarg])) {
         strcpy(outFileName, bang);
         strcat (outFileName, argv[(iarg++)+1]);
      }
      if (!strcmp("-plot", argv[iarg])) strcpy (plotFileName, argv[(iarg++)+1]);
      if (!strcmp("-ascii", argv[iarg])) strcpy (printFileName, argv[(iarg++)+1]);
      ++iarg;
   }
/*   
      open plot file
*/
   if (plotFileName[0] == 0) strcpy(plotFileName, "midiFoilPlot.ps/ps");
      else strcat(plotFileName, "/ps");
   cpgbeg(0, plotFileName, 1, 1);
/*
      check validity if output file
*/
   if (outFileName[0] == 0) {
      printf(" Need outFileName to proceed\n");
      Usage();
      return 1;
   }
/*
       if refDet specified, get detector def from that file
       else get detector def from input file
       get some head info and create output
*/
   if (refDetectorFileName[0] != 0) {
      fits_open_file (&inFile, refDetectorFileName, READONLY, &ierr);
      if (ierr != 0) {
         printf("fits openfile of refDet %s gave error %i \n",
            refDetectorFileName, ierr);
         printf("%s \n",strerror(errno));
         return 1;
      }
      outDetectorDef = read_imaging_detector_table(inFile, &ierr);
      i = fits_close_file(inFile, &ierr);
   }
/*
      what shall we use as input file
*/
   if (waveFileName[0] != 0) inFileName=waveFileName;
      else inFileName = curveFileName;
   if (!inFileName) {
      printf("Need at least one input file Name\n");
      Usage();
      return 1;
   }
   fits_open_file (&inFile, inFileName, READONLY, &ierr);
   if (ierr != 0) {
      printf("fits openfile of %s gave error %i \n",argv[1], ierr);
      printf("%s \n",strerror(errno));
      return 1;
   }
   i = fits_read_key(inFile, TSTRING, "HIERARCH ESO INS GRIS NAME", 
      grism, 0, &ierr);
   printf("disperser is = %s\n", grism); 
   if (refDetectorFileName[0] == 0) 
      outDetectorDef = read_imaging_detector_table(inFile, &ierr);
   idata          = open_imaging_data_table(inFile, &ierr);
/*
      need maxstep for output table definition
*/
   iKey = table_find_key(idata->def, "MAXSTEP"); 
   if (iKey>0)  table_get_keyword (idata->def, iKey, (void*)&maxstep, 
      &ierr);
/*
      create output data table 
*/

   fits_create_file (&outFile, outFileName, &ierr);
   if (ierr !=0) {
      fprintf (stderr, "error opening fits output %s %i\n", outFileName, ierr);
      return 1;
      }
/*    
      copy main header from input to output and close input
*/
   i = fits_movabs_hdu(inFile, 1, 0, &ierr);
   i = fits_copy_header(inFile, outFile, &ierr);
   i = fits_close_file(inFile, &ierr);
/*
      does user want a wavelength calibration?
*/
   if (waveFileName[0] !=0) {
      if (lineFileName[0] == 0) {
         printf("Wavelength calibration requires line file name\n");
         return  1;
      }
      if ((darkFileName[0] == 0) && (waveDarkName[0] ==0)) {
         printf("Wavelength calibration requires dark file name\n");
         return  1;
      }
      if (waveDarkName[0] == 0) strcpy(waveDarkName, darkFileName);
/*
      set line massaging parameters based on grism/prism
*/
   if (grism[0] == 'P' ) {
      sigma = 0.9;
      dx    =   3;
      order =   7;
   } else {
      sigma = 0.9; 
      dx    =   8;
      order =   4; 
   }
   printf("  %f = sigma\n %i = dx \n  %i = order\n ", sigma, dx, order);
/*
      read in and average foil data using minrtsFrameData utility
*/
   minrtsFrameDataFitsMeanRms(waveFileName, &meanData, &rms, &ierr); 

   minrtsFrameDataWriteFits(meanData, "meantest.fits", &ierr);
   minrtsFrameDataWriteFits(rms, "rmstest.fits", &ierr);
/*
      read in and average dark frame
*/
   minrtsFrameDataFitsMeanRms(waveDarkName, &meanDark, &rms, &ierr); 
   minrtsFrameDataWriteFits(meanDark, "meandark.fits", &ierr);
/*
      subtract dark from foil
*/
   minrtsFrameDataSub(meanData, meanData, meanDark);

   minrtsFrameDataWriteFits(meanData, "meandiff.fits", &ierr);
/*
      use rms frame data object as space for a mask to compress
      in y-direction
*/
   nregion = meanData->nregion;
   for (i=0; i<nregion; i++) {
      npix = rms->nx[i] * rms->ny[i];
      for (ipix=0; ipix<npix; ipix++) rms->data[i][ipix] = 1.;
   }
   compressed = minrtsFrameDataCompress (meanData, rms, 1); 

   nx = compressed->nx[0];
   work = (float *)malloc (nx * sizeof(float));
   for (i=0; i<4;i++) plotFreq[i] = (float *)malloc (nx * sizeof(float));
/*
      gaussian smooth
*/
   for (iregion =0; iregion<nregion; iregion++) {
      for (i=0;i<nx;i++) {
         np = 0.;
         work[i] = 0;
         for (j=-dx;j<=dx;j++) {
            k=i+j;
            if ((k >=0) && (k < nx)) {
               wt = exp(-0.5*j*j/(sigma*sigma));
               work[i] = work[i]+compressed->data[iregion][k] * wt;
               np = np + wt;
            }
         }
         if (np >0) work[i] = work[i]/np;
      }
      for (i=0;i<nx;i++) compressed->data[iregion][i] = work[i];
   }
/*
      second difference to supress slopes
*/
   for (iregion =0; iregion<nregion; iregion++) {
      nx = compressed->nx[iregion];
      for (i=1;i<nx-1;i++) work[i] = compressed->data[iregion][i] 
         - 0.5 *( compressed->data[iregion][i-1] +
           compressed->data[iregion][i+1]);
      work[0] = 0;
      work[nx-1] = 0;
      for (i=0;i<nx;i++) compressed->data[iregion][i] = work[i];
   }
/*
   normalize
*/
   for (iregion =0; iregion<nregion; iregion++) {
      nx = compressed->nx[iregion];
      wt = -1;
      for (i=0;i<nx;i++) {
         np = compressed->data[iregion][i];
         if (np > wt) wt = np;
         if (np < (-wt)) wt = -np;
      }
      for (i=0;i<nx;i++) compressed->data[iregion][i] = -compressed->data[iregion][i]/wt;
   }
   minrtsFrameDataWriteFits(compressed, "testcompr.fits", &ierr);
/*
      open file with  x-positions (1-relative for whole detector) and frequencies of line peaks
*/
   lineFile = fopen(lineFileName, "r");
   fscanf(lineFile, "%i", &nLine);
   printf("%i = nLine\n", nLine);
/*
      allocate arrays for line informatin
*/
   xLine    = (float *)malloc(nLine * sizeof(float));
   newXLine = (float *)malloc(nLine * 4 * sizeof(float));
   freqLine = (float *)malloc(nLine * sizeof(float));
   lWeight  = (float *)malloc(nLine * sizeof(float));
   xCoord   = (float *)malloc(nx * sizeof(float));
/*
      reference pixel is half way down array
*/
   crpix = nx/2;
   printf("%f = crpix \n",crpix);
/*
      read in x-pos and freq and subtract "corner" so positions are
      0-relative within window
*/
   for (i=0;i<nLine;i++) fscanf(lineFile, "%f %f ",xLine+i, freqLine+i);
   for (i=0;i<nLine;i++) xLine[i] = xLine[i] - meanData->x0[0];
/*
      independent variable for line peaks is relative to crpix
*/
   for (i=0;i<nx;i++) xCoord[i] = 1 + (float)i - crpix; /* 1-rel */
   for (i=0;i<nLine;i++) lWeight[i] = 1.0;
/*    fit peaks */
   for (iregion=0; iregion<nregion; iregion++) {
      for (i=0; i<nLine;i++) {
         k = (int)(xLine[i] - dx);
         newXLine[i+nLine*iregion] = minrtsQPk(2*dx+1, xCoord+k, 
            compressed->data[iregion]+k, &x, 0);  
      }
/*
      fit polynomial to peak positions
*/
      minrtsPolyFit(order, nLine, newXLine+nLine*iregion, freqLine, lWeight,
         polyCoeff);
for (i=0;i<nLine;i++) printf("%f\n", newXLine[i+iregion*nLine]);
for (i=0;i<=order;i++) printf("%f\n", polyCoeff[i]);
/*
      compute complete set of frequencies for plotting later
*/
      for (i=0;i<nx;i++) plotFreq[iregion][i] = 0.;
      for (i=0;i<nx;i++) for (j=0;j<=order;j++) {
         plotFreq[iregion][i] =
         plotFreq[iregion][i] + polyCoeff[j]* pow(xCoord[i],j);
      }
/*
      format polynomial coefficients into CD matrix and distortion
      coefficients
*/
      outDetectorDef->crpix[iregion][0] = crpix;
      outDetectorDef->crval[iregion][0] = polyCoeff[0];
      outDetectorDef->cd[iregion][0] = polyCoeff[1];
      outDetectorDef->cd[iregion][1] = 0.;
      outDetectorDef->cd[iregion][2] = 0.;
/*
           initialize dmp,dmc to zero
*/
      for (i=0;i<outDetectorDef->max_coef;i++) {
         outDetectorDef->dmp[iregion][i][0][0] = 0;
         outDetectorDef->dmp[iregion][i][1][0] = 0;
         outDetectorDef->dmc[iregion][i][0]    = 0.;
      }
/*            set x-exponents to 1,2,3...; */
      for (i=0; i<order; i++) outDetectorDef->dmp[iregion][i][0][0] = 1+i;
         outDetectorDef->dmp[iregion][0][1][1] = 1;
/*            write coefficients into correct place */
      for (i=0; i<order; i++) outDetectorDef->dmc[iregion][i][0] =
         polyCoeff[1+i]/polyCoeff[1];
         outDetectorDef->dmc[iregion][0][1] = 1;
   } /* iRegion */
/* 
      make plots
*/
   for (i=0; i<nx; i++) work[i] = 1.;
   cpgscr(1,1.0,0.,0.);
   cpgscr(2,0.,1.0,0.);
   cpgscr(3,0.,0.,1.0);
   cpgscr(4,1.0,0.,1.0);
   cpgsvp(.05, .95, 0.1, .95);
/*   cpgsubp(1,4); */
   for (iregion=0; iregion<nregion; iregion++) {
      cpgswin(xCoord[0], xCoord[nx-1], -1., 1.);
      cpgbox("ABCTN", 0., 0., "ABTN", 0., 0);
      cpglab ("x", "flux"," ");
      cpgline(nx, xCoord, compressed->data[iregion]);
      for (i=0;i<nLine;i++) {
         work[0] = newXLine[i+iregion*nLine];
         work[1] = work[0];
         work[2] = -1.;
         work[3] = 1.;
         cpgline(2, work, work+2);
      }
/*      cpgpt(nLine, newXLine+iregion*nLine, work, 17); */
      cpgpage();
   }
/*   cpgsubp(1,1); */
   cpgswin(2.0e13, 4.5e13, -1., 1.);
   cpgbox("ABCTN", 0., 0., "ABCTN", 0., 0);
   cpglab ("x", "flux"," ");
   for (iregion=0; iregion<nregion; iregion++) {
      cpgsci(iregion+1);
      cpgline(nx, plotFreq[iregion], compressed->data[iregion]);
   }
   cpgsci(0);
   for (i=0;i<nLine;i++) {
      work[0] = freqLine[i];
      work[1] = work[0];
      work[2] = -1.;
      work[3] = 1.;
      cpgline(2, work, work+2);
   }
   cpgpage();
   cpgsvp(.05, .95, 0.1, .95);
   cpgswin(xCoord[0], xCoord[nx-1], 2.e13, 4.5e13);
   cpgbox("ABCTN", 0., 0., "ABTN", 0., 0);
   cpglab ("x", "freq"," ");
   for (iregion=0; iregion<nregion; iregion++) {
      cpgsci(iregion+1);
      cpgline(nx, xCoord, plotFreq[iregion]);
   }
   cpgpage();
/*
       make ascii printout if requested
*/
   if (plotFileName[0] != 0) {
      printFile = fopen(printFileName, "w");
      for (iregion=0;iregion<nregion;iregion++){
         fprintf(printFile, "Window %s \n\n",iregion+1);
         for (i=0;i<nx;i++) fprintf(printFile," %i %f\n", xCoord[i],
            cLight/plotFreq[iregion][i]);
         fprintf(printFile,"\n");
      }
   }
/*
      clean up
*/
   minrtsFrameDataKill(meanData);
   minrtsFrameDataKill(meanDark);
   minrtsFrameDataKill(rms);
   minrtsFrameDataKill(compressed);
   free(work);
   for (i=0;i<4;i++) free(plotFreq[i]);
   free(xLine);
   free(newXLine);
   free(freqLine);
   free(lWeight);
   free(xCoord);
   fclose(lineFile);
   }  /* end of do we do wavelength calibration */
/*
      do we do curvature calibration?
*/
   if (curveFileName[0] !=0) {
      if ((darkFileName[0] == 0) && (curveDarkName[0] ==0)) {
         printf("Wavelength calibration requires dark file name\n");
         return  1;
      }
      if (curveDarkName[0] == 0) strcpy(curveDarkName, darkFileName);
/*
      set order based on grism
*/
   if (grism[0] == 'P' ) order =   3;
      else order =   4; 
/*
      read in and average foil data using minrtsFrameData utility
*/
   minrtsFrameDataFitsMeanRms(curveFileName, &meanData, &rms, &ierr); 

   minrtsFrameDataWriteFits(meanData, "curvemeantest.fits", &ierr);
   minrtsFrameDataWriteFits(rms, "curvermstest.fits", &ierr);
/*
      read in and average dark frame
*/
   minrtsFrameDataFitsMeanRms(curveDarkName, &meanDark, &rms, &ierr); 
   minrtsFrameDataWriteFits(meanDark, "curvemeandark.fits", &ierr);
/*
      subtract dark from foil
*/
   minrtsFrameDataSub(meanData, meanData, meanDark);

   minrtsFrameDataWriteFits(meanData, "curvemeandiff.fits", &ierr);
   nregion = meanData->nregion;
   nx      = meanData->nx[0];
   work     = (float *)malloc(nx * sizeof(float));
   xCoord   = (float *)malloc(nx * sizeof(float));
   lWeight  = (float *)malloc(nx * sizeof(float));
   for (iregion=0;iregion<4;iregion++) {
      plotCurve[iregion] = (float *)malloc(nx * sizeof(float));
      curveData[iregion] = (float *)malloc(nx * sizeof(float));
   }
   crpix = nx/2;
   for (i=0;i<nx;i++) xCoord[i] = 1 + (float)i - crpix; /* 1-rel x-coordinates */
   for (iregion=0; iregion<nregion;iregion++) {
      ny = meanData->ny[iregion];
      for (j=0;j<ny;j++) work[ny+j] = j+1; /* y-coordinates */
/*
      loop over x-points, 
      copy column values into work array for minrtsQPk routine
      sum pixel values in column as weight
*/
      for (i=0;i<nx;i++) {
         lWeight[i] = 0;
         for (j=0;j<ny;j++) {
            work[j] = meanData->data[iregion][i+j*nx];
            lWeight[i] = lWeight[i] + work[j];
         }
/*
      put position of peak in column -> curveData
*/
         if (lWeight[i] <0) lWeight[i] = 0;
         curveData[iregion][i] = minrtsQPk(ny, work+ny, work, &x, 0);
      }  /* nx */
/*
      fit polynomial to peak positions
      and store fit for plotting
*/
      minrtsPolyFit(order, nx, xCoord, curveData[iregion], lWeight,
         polyCoeff);
for (i=0;i<=order;i++) printf("%f\n", polyCoeff[i]);
      for (i=0;i<nx;i++) plotCurve[iregion][i] = 0.;
      for (i=0;i<nx;i++) for (j=0;j<=order;j++) {
         plotCurve[iregion][i] =
         plotCurve[iregion][i] + polyCoeff[j]* pow(xCoord[i],j);
      }
/*
      insert solutions into detector tables
*/
      outDetectorDef->crpix[iregion][1] = (int)ny/2;
      outDetectorDef->crval[iregion][1] = ny*0.5-polyCoeff[0];
/*
      zero dmp/dmc for output y-variable
*/
      for (i=0;i<outDetectorDef->max_coef;i++) {
         outDetectorDef->dmp[iregion][i][0][1] = 0;
         outDetectorDef->dmp[iregion][i][1][1] = 0;
         outDetectorDef->dmc[iregion][i][1]    = 0.;
      }
/*
      put default linear y-variation in 1st position
*/
      outDetectorDef->cd[iregion][3] = 1.;
      outDetectorDef->dmp[iregion][0][1][1] = 1;
      outDetectorDef->dmp[iregion][0][0][1] = 0;
      outDetectorDef->dmc[iregion][0][1] = 1.;
/*            set x-exponents to 1,2,3...; */
      for (i=0; i<order; i++) outDetectorDef->dmp[iregion][i+1][0][1] = 1+i;
/*            write coefficients into correct place */
      for (i=0; i<order; i++) outDetectorDef->dmc[iregion][i+1][1] =
         -polyCoeff[1+i];

   } /* end of iregion */
   minrtsFrameDataKill(meanData);
   minrtsFrameDataKill(meanDark);
   minrtsFrameDataKill(rms);
   free(work);
   free(xCoord);
   free(lWeight);
   for (iregion=0;iregion<4;iregion++) {
      free(plotCurve[iregion]);
      free(curveData[iregion]);
   }
   } /* end of curvature calibration */
/*
      write output detector table
      create output data table (but no data rows)
*/
   odata = make_imaging_data_table_def (outFile,
      outDetectorDef, maxstep, &ierr);
   if (ierr !=0) {
      fprintf (stderr, "error creating data definition\n");
      return 1;
   }

/*
      close output stuff
*/
   i = close_imaging_data_table(outFile, odata, &ierr);
   cpgend();

   return 0;
}
